import mmh3
import numpy as np
from tqdm.auto import tqdm
from typing import List, Optional, Dict, Union, Tuple
from collections import Counter
from chn_tokenizer import CHNTokenizer


class BM25Encoder():
    def __init__(self, b = 0.75, k1 = 1.2, language = "chn"):
        self.b = b
        self.k1 = k1
        self._tokenizer = CHNTokenizer()
        if language == "eng":
            self._tokenizer = ENGTokenizer()

        self.doc_freq = {}
        self.n_docs = 0
        self.avgdl = 0
        self.sum_doc_len = 0
        self.doc_freq_counter = Counter()

    def fit(self, corpus):
        """
        在全量数据上进行fit，获取一些统计信息；例如平均文档长度等。
        """
        for doc in tqdm(corpus):
            if not isinstance(doc, str):
                raise ValueError("Must be fit on a list of strings")
            indices, tf = self._tf(doc)
            if len(indices) == 0:
                continue
            self.n_docs += 1
            self.sum_doc_len += sum(tf)
            self.doc_freq_counter.update(indices)

        self.doc_freq = dict(self.doc_freq_counter)
        self.avgdl = self.sum_doc_len / self.n_docs
        return self

    def encode_documents(self, texts):
        if self.doc_freq is None or self.n_docs is None or self.avgdl is None:
            raise ValueError("BM25 must be fit before encoding documents")

        if isinstance(texts, str):
            return self._encode_one_document(texts)
        elif isinstance(texts, list):
            return [self._encode_one_document(text) for text in texts]
        else:
            raise ValueError("Only support string or list[string]")

    def _encode_one_document(self, text):
        indices, doc_tf = self._tf(text)
        tf = np.array(doc_tf)
        tf_sum = sum(tf)

        tf_normed = tf / (
            self.k1 * (1.0 - self.b + self.b * (tf_sum / self.avgdl)) + tf
        )
        zipped = zip(indices, tf_normed.tolist())
        sorted_data = sorted(zipped)
        indices, values = zip(*sorted_data)
        return {
            "indices": list(indices),
            "values": list(values),
        }

    def encode_queries(self, texts):
        if self.doc_freq is None or self.n_docs is None or self.avgdl is None:
            raise ValueError("BM25 must be fit before encoding queries")

        if isinstance(texts, str):
            return self._encode_one_query(texts)
        elif isinstance(texts, list):
            return [self._encode_one_query(text) for text in texts]
        else:
            raise ValueError("Only support string or list[string]")

    def _encode_one_query(self, text):
        indices, query_tf = self._tf(text)

        tf = np.array([self.doc_freq.get(idx, 1) for idx in indices])  # type: ignore
        idf = np.log((self.n_docs + 1) / (tf + 0.5))  # type: ignore
        idf_normed = idf / idf.sum()
        zipped = zip(indices, idf_normed.tolist())
        sorted_data = sorted(zipped)
        indices, values = zip(*sorted_data)
        return {
            "indices": list(indices),
            "values": list(values),
        }

    @staticmethod
    def _token_id(token):
        """使用hash值作为token id"""
        return mmh3.hash(token, signed=False)

    def _tf(self, text):
        """
        计算一段文本的tf
        """
        counts = Counter((self._token_id(token) for token in self._tokenizer(text)))

        items = list(counts.items())
        return [idx for idx, _ in items], [val for _, val in items]
